// Copyright 2016 Chris Drake
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.


#include "boolexpr/boolexpr.h"


using std::make_shared;


namespace boolexpr {


zero_t
zero()
{
    static auto _zero = make_shared<Zero>();
    return _zero;
}


one_t
one()
{
    static auto _one = make_shared<One>();
    return _one;
}


log_t
logical()
{
    static auto _log = make_shared<Logical>();
    return _log;
}


ill_t
illogical()
{
    static auto _ill = make_shared<Illogical>();
    return _ill;
}


}  // namespace boolexpr
